# noqa

"""
A file-based database, requires a writeable directory (path), which is usually
a requirement for drivers. It can store any additional metadata, accessible  as
a dictionary and serialized to and from JSON. This metadata file gets created
if it doesn't exist.
"""
from __future__ import annotations

import json
import os


def connection(path, serializer="raw"):
    serializers = {"raw": RawSerializer, "json": JSONSerializer}
    return FileBasedDatabase(path, serializer=serializers[serializer])


class FileBasedDatabase:
    def __init__(self, directory_path, serializer=None):
        self.directory_path = directory_path
        self.serializer = serializer or RawSerializer
        self.files = []

    def get(self, name):
        """
        Find the file from name, return the serializer instantiated with it. If
        the file path does not exist it will not return anything. Similar to
        how a `.get()` would work in a dictionary.
        """
        if not name.endswith(self.serializer.ext):
            name = f"{name}{self.serializer.ext}"
        if self.files == []:
            self._update_file_cache()
        if name in self.files:
            path = os.path.join(self.directory_path, name)
            return self.serializer(path)
        return None

    def create(self, name):
        """
        If `name` does not exist, go ahead and return a serializer with the
        path regardless, so that it can commit data to it. It does not check if
        the file exists or not, the serializer should be able to write to it.
        """
        if not name.endswith(self.serializer.ext):
            name = f"{name}{self.serializer.ext}"
        path = os.path.join(self.directory_path, name)
        return self.serializer(path)

    def get_all(self):
        """
        Retrieve all the files in the database directory path except for the
        metadata file which is special and not part of the saved data.
        """
        for name in self._update_file_cache():
            yield self.serializer(os.path.join(self.directory_path, name))

    def _update_file_cache(self):
        all_files = [i for i in os.listdir(self.directory_path) if ".__meta__.json" not in i]
        self.files = [i for i in all_files if i.endswith(self.serializer.ext)]
        return self.files

    def get_metadata(self):
        """
        This is a special method that loads (or creates) the metadata file in
        the directory. This metadata file allows to stick any information as
        a flat JSON file that can be loaded and saved as a regular file.

        If the file does not exist, it gets initialized with an empty JSON
        object so that callers don't get exceptions.
        """
        name = ".__meta__.json"
        path = os.path.join(self.directory_path, name)
        if not os.path.exists(path):
            with open(path, "w") as fp:
                fp.write("{}")
        meta = JSONSerializer(path)
        meta.load()
        return meta


class JSONSerializer:
    ext = ".json"

    def __init__(self, path):
        self.path = path
        self.data = {}

    def load(self):
        with open(self.path) as fp:
            self.data = json.load(fp)
        return self.data

    def commit(self, data=None):
        """
        `data` is expected as a Python dictionary, this method takes care of
        serializing back to JSON before saving
        """
        # allow to commit what exists already as self.data if data is not passed in
        data = data or self.data
        self.data.update(data)
        with open(self.path, "w") as fp:
            json.dump(data, fp)


class RawSerializer:
    ext = ".txt"

    def __init__(self, path):
        self.path = path
        self.data = None

    def load(self):
        with open(self.path) as fp:
            self.data = fp.read()
        return self.data

    def commit(self, data):
        with open(self.path, "w") as fp:
            fp.write(data)



from __future__ import annotations

import os
from typing import TYPE_CHECKING, Any

import requests
from lxml import etree
from xsdata.formats.dataclass.parsers import XmlParser
from xsdata.formats.dataclass.parsers.config import ParserConfig

from vunnel.providers.mariner.model import Definition, RpminfoObject, RpminfoState, RpminfoTest
from vunnel.utils import retry_with_backoff
from vunnel.utils.vulnerability import FixedIn, Vulnerability

if TYPE_CHECKING:
    import logging
    from collections.abc import Generator

    from vunnel.workspace import Workspace

LESS_THAN_OR_EQUAL_TO = "less than or equal"

IGNORED_PATCHABLE_VALUES = ["Not Applicable"]


class MarinerXmlFile:
    def __init__(self, oval_file_path: str, logger: logging.Logger):
        parser_config = ParserConfig(
            fail_on_converter_warnings=False,
            fail_on_unknown_attributes=False,
            fail_on_unknown_properties=False,
        )
        xml_parser = XmlParser(config=parser_config)
        root = etree.parse(oval_file_path)
        nsmap = etree.XPath("/*")(root)[0].nsmap
        default = nsmap[None]
        nsmap["default"] = default
        del nsmap[None]
        self.logger = logger

        # Go element by element to have flexible error handling
        self.definitions = []
        for def_element in etree.XPath("//default:definition", namespaces=nsmap)(root):
            try:
                element_bytes = etree.tostring(def_element)
                definition = xml_parser.from_string(element_bytes.decode(), Definition)
                self.definitions.append(definition)
            except Exception as ex:
                self.logger.warning(f"skipping definition element in {oval_file_path} due to {ex}")
                pass

        self.tests_by_id = {}
        for test_element in etree.XPath("//linux-def:rpminfo_test", namespaces=nsmap)(root):
            try:
                element_bytes = etree.tostring(test_element)
                test_obj = xml_parser.from_string(element_bytes.decode(), RpminfoTest)
                self.tests_by_id[test_obj.id] = test_obj
            except Exception as ex:
                self.logger.warning(f"skipping rpminfo_test element in {oval_file_path} due to {ex}")
        self.objects_by_id = {}
        for obj_element in etree.XPath("//linux-def:rpminfo_object", namespaces=nsmap)(root):
            try:
                element_bytes = etree.tostring(obj_element)
                obj = xml_parser.from_string(element_bytes.decode(), RpminfoObject)
                self.objects_by_id[obj.id] = obj
            except Exception as ex:
                self.logger.warning(f"skipping rpminfo_object element in {oval_file_path} due to {ex}")

        self.states_by_id = {}
        for state_element in etree.XPath("//linux-def:rpminfo_state", namespaces=nsmap)(root):
            try:
                element_bytes = etree.tostring(state_element)
                state_obj = xml_parser.from_string(element_bytes.decode(), RpminfoState)
                self.states_by_id[state_obj.id] = state_obj
            except Exception as ex:
                self.logger.warning(f"skipping rpminfo_object element in {oval_file_path} due to {ex}")
        self.mariner_version = oval_file_path.split("-")[-2]

    def name_and_version(self, test_id: str) -> tuple[str | None, str | None]:
        test = self.tests_by_id.get(test_id, None)
        if test is None or test.object_value is None or test.state is None:
            return (None, None)
        obj = self.objects_by_id.get(test.object_value.object_ref, None)
        if obj is None:
            return (None, None)
        state = self.states_by_id.get(test.state.state_ref, None)
        if state is None or state.evr is None:
            return (None, None)
        return (obj.name, state.evr.value)

    def namespace_name(self) -> str:
        return f"mariner:{self.mariner_version}"

    def get_test(self, definition: Definition) -> RpminfoTest | None:
        if definition is None or definition.criteria is None or definition.criteria.criterion is None:
            return None
        return self.tests_by_id.get(definition.criteria.criterion.test_ref, None)

    def get_state(self, definition: Definition) -> RpminfoState | None:
        test = self.get_test(definition)
        if test is None or test.state is None or test.state.state_ref is None:
            return None
        return self.states_by_id.get(test.state.state_ref, None)

    def get_object(self, definition: Definition) -> RpminfoObject | None:
        test = self.get_test(definition)
        if test is None or test.object_value is None or test.object_value.object_ref is None:
            return None
        return self.objects_by_id.get(test.object_value.object_ref, None)

    def make_fixed_in(self, definition: Definition) -> FixedIn | None:
        state = self.get_state(definition)
        obj = self.get_object(definition)
        if state is None or state.evr is None:
            return None
        if obj is None or obj.name is None:
            return None
        version = state.evr.value
        # There are 2 choices for state.ever.operation: "less than" or "less than or equal to".
        # So for this vulnerability, either the statement, "versions < 3.2.1 are vulernable"
        # or the statement "versions <= 3.2.1 are vulnerable". In the second statement,
        # the data has no information about any fixed version, so we report "None"
        # as the fixed version, meaning we consider all version vulnerable.
        # For example, if version 3.2.1 of a library is vulnerable, and is the latest version
        # mariner data might have "versions <= 3.2.1" is vulnerable.
        if state.evr.operation == LESS_THAN_OR_EQUAL_TO:
            version = "None"  # legacy API needs the string "None" instead of None
        return FixedIn(
            Name=obj.name,
            NamespaceName=self.namespace_name(),
            VersionFormat="rpm",
            Version=version,
            Module=None,
            VendorAdvisory=None,
        )

    def vulnerability_id(self, definition: Definition) -> str | None:
        if definition.metadata is None or definition.metadata.reference is None or definition.metadata.reference.ref_id is None:
            return None
        return definition.metadata.reference.ref_id

    def description(self, definition: Definition) -> str | None:
        if not definition.metadata:
            return None
        return definition.metadata.description

    def vulnerabilities(self) -> Generator[Vulnerability, None, None]:
        for d in self.definitions:
            if d.metadata is None or d.metadata.severity is None:
                self.logger.warning("skipping definition because severity could not be found")
                continue
            if d.metadata and d.metadata.patchable and d.metadata.patchable in IGNORED_PATCHABLE_VALUES:
                continue
            link = ""
            if d.metadata.reference and d.metadata.reference.ref_url:
                link = d.metadata.reference.ref_url
            fixed_in = self.make_fixed_in(d)
            if not fixed_in:
                continue
            vulnerability_id = self.vulnerability_id(d)
            if not vulnerability_id:
                continue
            yield Vulnerability(
                Name=vulnerability_id,  # intentional; legacy API uses Name as field for vulnerability ID.
                NamespaceName=self.namespace_name(),
                Description=self.description(d) or "",
                Severity=d.metadata.severity,
                Link=link,
                CVSS=[],
                FixedIn=[fixed_in],
                Metadata={},  # Ideally, there's no metadata here.
            )


MARINER_URL_BASE = "https://raw.githubusercontent.com/microsoft/CBL-MarinerVulnerabilityData/main/{}"
MARINER_URL_FILENAME = "cbl-mariner-{}-oval.xml"


class Parser:
    def __init__(self, workspace: Workspace, download_timeout: int, allow_versions: list[Any], logger: logging.Logger):
        self.workspace = workspace
        self.download_timeout = download_timeout
        self.allow_versions = allow_versions
        self._urls: set[str] = set()
        self.logger = logger

    def _download(self) -> list[str]:
        return [self._download_version(v) for v in self.allow_versions]

    @retry_with_backoff()
    def _download_version(self, version: str) -> str:
        filename = MARINER_URL_FILENAME.format(version)
        url = MARINER_URL_BASE.format(filename)
        r = requests.get(url, timeout=self.download_timeout)
        destination = os.path.join(self.workspace.input_path, filename)
        with open(destination, "wb") as writer:
            writer.write(r.content)
        self._urls.add(url)
        return destination

    @property
    def urls(self) -> list[str]:
        return list(self._urls)

    def get(self) -> Generator[tuple[str, str, dict[str, dict[str, Any]]], None, None]:
        for oval_file_path in self._download():
            parsed_file = MarinerXmlFile(oval_file_path, self.logger)
            for v in parsed_file.vulnerabilities():
                yield v.NamespaceName, v.Name, v.to_payload()



from __future__ import annotations

import os
from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from vunnel import provider, result, schema
from vunnel.providers.nvd.manager import Manager

if TYPE_CHECKING:
    import datetime


@dataclass
class Config:
    runtime: provider.RuntimeConfig = field(
        default_factory=lambda: provider.RuntimeConfig(
            result_store=result.StoreStrategy.SQLITE,
            existing_results=result.ResultStatePolicy.KEEP,
        ),
    )
    request_timeout: int = 125
    api_key: str = "env:NVD_API_KEY"

    def __post_init__(self) -> None:
        if self.api_key.startswith("env:"):
            self.api_key = os.environ.get(self.api_key[4:], "")

    def __str__(self) -> str:
        # sanitize secrets from any output
        api_value = self.api_key
        str_value = super().__str__()
        if not api_value:
            return str_value
        return str_value.replace(api_value, "********")


class Provider(provider.Provider):
    def __init__(self, root: str, config: Config | None = None):
        if not config:
            config = Config()
        super().__init__(root, runtime_cfg=config.runtime)
        self.config = config

        self.logger.debug(f"config: {config}")

        if self.config.runtime.skip_if_exists and config.runtime.existing_results != result.ResultStatePolicy.KEEP:
            raise ValueError(
                "if 'skip_if_exists' is set then 'runtime.existing_results' must be 'keep' "
                "(otherwise incremental updates will fail)",
            )

        self.schema = schema.NVDSchema()
        self.manager = Manager(
            workspace=self.workspace,
            download_timeout=self.config.request_timeout,
            api_key=self.config.api_key,
            logger=self.logger,
        )

    @classmethod
    def name(cls) -> str:
        return "nvd"

    def update(self, last_updated: datetime.datetime | None) -> tuple[list[str], int]:
        with self.results_writer() as writer:
            for identifier, record in self.manager.get(
                skip_if_exists=self.config.runtime.skip_if_exists,
                last_updated=last_updated,
            ):
                writer.write(
                    identifier=identifier.lower(),
                    schema=self.schema,
                    payload=record,
                )

        return self.manager.urls, len(writer)



from dataclasses import dataclass, field
from typing import List, Optional, Union
from xsdata.models.datatype import XmlDateTime


@dataclass
class Evr:
    class Meta:
        name = "evr"
        namespace = "http://oval.mitre.org/XMLSchema/oval-definitions-5#linux"

    datatype: Optional[str] = field(
        default=None,
        metadata={
            "type": "Attribute",
        }
    )
    operation: Optional[str] = field(
        default=None,
        metadata={
            "type": "Attribute",
        }
    )
    value: str = field(
        default=""
    )


@dataclass
class Object:
    class Meta:
        name = "object"
        namespace = "http://oval.mitre.org/XMLSchema/oval-definitions-5#linux"

    object_ref: Optional[str] = field(
        default=None,
        metadata={
            "type": "Attribute",
        }
    )


@dataclass
class RpminfoObject:
    class Meta:
        name = "rpminfo_object"
        namespace = "http://oval.mitre.org/XMLSchema/oval-definitions-5#linux"

    id: Optional[str] = field(
        default=None,
        metadata={
            "type": "Attribute",
        }
    )
    version: Optional[int] = field(
        default=None,
        metadata={
            "type": "Attribute",
        }
    )
    name: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
        }
    )


@dataclass
class State:
    class Meta:
        name = "state"
        namespace = "http://oval.mitre.org/XMLSchema/oval-definitions-5#linux"

    state_ref: Optional[str] = field(
        default=None,
        metadata={
            "type": "Attribute",
        }
    )


@dataclass
class Affected:
    class Meta:
        name = "affected"
        namespace = "http://oval.mitre.org/XMLSchema/oval-definitions-5"

    family: Optional[str] = field(
        default=None,
        metadata={
            "type": "Attribute",
        }
    )
    platform: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
        }
    )


@dataclass
class Criterion:
    class Meta:
        name = "criterion"
        namespace = "http://oval.mitre.org/XMLSchema/oval-definitions-5"

    comment: Optional[str] = field(
        default=None,
        metadata={
            "type": "Attribute",
        }
    )
    test_ref: Optional[str] = field(
        default=None,
        metadata={
            "type": "Attribute",
        }
    )


@dataclass
class Generator:
    class Meta:
        name = "generator"
        namespace = "http://oval.mitre.org/XMLSchema/oval-definitions-5"

    product_name: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://oval.mitre.org/XMLSchema/oval-common-5",
        }
    )
    product_version: Optional[int] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://oval.mitre.org/XMLSchema/oval-common-5",
        }
    )
    schema_version: Optional[float] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://oval.mitre.org/XMLSchema/oval-common-5",
        }
    )
    timestamp: Optional[XmlDateTime] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://oval.mitre.org/XMLSchema/oval-common-5",
        }
    )
    content_version: Optional[int] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "http://oval.mitre.org/XMLSchema/oval-common-5",
        }
    )


@dataclass
class Reference:
    class Meta:
        name = "reference"
        namespace = "http://oval.mitre.org/XMLSchema/oval-definitions-5"

    ref_id: Optional[str] = field(
        default=None,
        metadata={
            "type": "Attribute",
        }
    )
    ref_url: Optional[str] = field(
        default=None,
        metadata={
            "type": "Attribute",
        }
    )
    source: Optional[str] = field(
        default=None,
        metadata={
            "type": "Attribute",
        }
    )


@dataclass
class RpminfoState:
    class Meta:
        name = "rpminfo_state"
        namespace = "http://oval.mitre.org/XMLSchema/oval-definitions-5#linux"

    id: Optional[str] = field(
        default=None,
        metadata={
            "type": "Attribute",
        }
    )
    version: Optional[int] = field(
        default=None,
        metadata={
            "type": "Attribute",
        }
    )
    evr: Optional[Evr] = field(
        default=None,
        metadata={
            "type": "Element",
        }
    )


@dataclass
class RpminfoTest:
    class Meta:
        name = "rpminfo_test"
        namespace = "http://oval.mitre.org/XMLSchema/oval-definitions-5#linux"

    check: Optional[str] = field(
        default=None,
        metadata={
            "type": "Attribute",
        }
    )
    comment: Optional[str] = field(
        default=None,
        metadata={
            "type": "Attribute",
        }
    )
    id: Optional[str] = field(
        default=None,
        metadata={
            "type": "Attribute",
        }
    )
    version: Optional[int] = field(
        default=None,
        metadata={
            "type": "Attribute",
        }
    )
    object_value: Optional[Object] = field(
        default=None,
        metadata={
            "name": "object",
            "type": "Element",
        }
    )
    state: Optional[State] = field(
        default=None,
        metadata={
            "type": "Element",
        }
    )


@dataclass
class Criteria:
    class Meta:
        name = "criteria"
        namespace = "http://oval.mitre.org/XMLSchema/oval-definitions-5"

    operator: Optional[str] = field(
        default=None,
        metadata={
            "type": "Attribute",
        }
    )
    criterion: Optional[Criterion] = field(
        default=None,
        metadata={
            "type": "Element",
        }
    )


@dataclass
class Metadata:
    class Meta:
        name = "metadata"
        namespace = "http://oval.mitre.org/XMLSchema/oval-definitions-5"

    title: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
        }
    )
    affected: Optional[Affected] = field(
        default=None,
        metadata={
            "type": "Element",
        }
    )
    reference: Optional[Reference] = field(
        default=None,
        metadata={
            "type": "Element",
        }
    )
    patchable: Optional[Union[bool, str]] = field(
        default=None,
        metadata={
            "type": "Element",
        }
    )
    advisory_date: Optional[XmlDateTime] = field(
        default=None,
        metadata={
            "type": "Element",
        }
    )
    advisory_id: Optional[Union[str, int]] = field(
        default=None,
        metadata={
            "type": "Element",
        }
    )
    severity: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
        }
    )
    description: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
        }
    )


@dataclass
class Objects:
    class Meta:
        name = "objects"
        namespace = "http://oval.mitre.org/XMLSchema/oval-definitions-5"

    rpminfo_object: List[RpminfoObject] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "http://oval.mitre.org/XMLSchema/oval-definitions-5#linux",
        }
    )


@dataclass
class Definition:
    class Meta:
        name = "definition"
        namespace = "http://oval.mitre.org/XMLSchema/oval-definitions-5"

    class_value: Optional[str] = field(
        default=None,
        metadata={
            "name": "class",
            "type": "Attribute",
        }
    )
    id: Optional[str] = field(
        default=None,
        metadata={
            "type": "Attribute",
        }
    )
    version: Optional[int] = field(
        default=None,
        metadata={
            "type": "Attribute",
        }
    )
    metadata: Optional[Metadata] = field(
        default=None,
        metadata={
            "type": "Element",
        }
    )
    criteria: Optional[Criteria] = field(
        default=None,
        metadata={
            "type": "Element",
        }
    )


@dataclass
class States:
    class Meta:
        name = "states"
        namespace = "http://oval.mitre.org/XMLSchema/oval-definitions-5"

    rpminfo_state: List[RpminfoState] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "http://oval.mitre.org/XMLSchema/oval-definitions-5#linux",
        }
    )


@dataclass
class Tests:
    class Meta:
        name = "tests"
        namespace = "http://oval.mitre.org/XMLSchema/oval-definitions-5"

    rpminfo_test: List[RpminfoTest] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "http://oval.mitre.org/XMLSchema/oval-definitions-5#linux",
        }
    )


@dataclass
class Definitions:
    class Meta:
        name = "definitions"
        namespace = "http://oval.mitre.org/XMLSchema/oval-definitions-5"

    definition: List[Definition] = field(
        default_factory=list,
        metadata={
            "type": "Element",
        }
    )


@dataclass
class OvalDefinitions:
    class Meta:
        name = "oval_definitions"
        namespace = "http://oval.mitre.org/XMLSchema/oval-definitions-5"

    schema_location: Optional[str] = field(
        default=None,
        metadata={
            "name": "schemaLocation",
            "type": "Attribute",
            "namespace": "http://www.w3.org/2001/XMLSchema-instance",
        }
    )
    generator: Optional[Generator] = field(
        default=None,
        metadata={
            "type": "Element",
        }
    )
    definitions: Optional[Definitions] = field(
        default=None,
        metadata={
            "type": "Element",
        }
    )
    tests: Optional[Tests] = field(
        default=None,
        metadata={
            "type": "Element",
        }
    )
    objects: Optional[Objects] = field(
        default=None,
        metadata={
            "type": "Element",
        }
    )
    states: Optional[States] = field(
        default=None,
        metadata={
            "type": "Element",
        }
    )



# flake8: noqa

"""
A generic framework for parsing an OVAL xml file. Design is based on separate collections/sections in an OVAL schema.
Each section is associated with a parser that can be overridden by the driver. Parsed output represents a view of the
OVAL content, it's up to the driver to transform it into normalized feed data
"""
from __future__ import annotations

import enum
import gzip
import logging
import os
import re
import xml.etree.ElementTree as ET  # nosec (this is only used to get the definition for Element, which is not in defusedxml)
from abc import ABC, abstractmethod
from collections import defaultdict
from dataclasses import dataclass

from defusedxml.ElementTree import iterparse


class OVALElementEnum(enum.Enum):
    """
    XML tags of OVAL collection items
    """

    VULNERABILITY = "definition"
    TEST = "rpminfo_test"
    ARTIFACT = "rpminfo_object"
    VERSION = "rpminfo_state"


@dataclass
class OVALParserConfig:
    """
    Configuration for parsing an OVAL file. While the OVAL format is standard, each publisher expresses data slightly
    differently. Use an instance of this class to pass the configuration specific to an OVAL or a publisher
    """

    # to be defined
    # regexes
    platform_regex: re.Pattern
    artifact_regex: re.Pattern

    # xpath queries
    source_url_xpath_query: str

    # severity
    severity_map: dict

    # defaults
    # regexes
    tag_regex = re.compile(r"\{http://oval.mitre.org/XMLSchema/.*\}(\w*)")
    namespace_regex = re.compile(r"(\{http://oval.mitre.org/XMLSchema/.*\})\w*")

    # xpath queries
    title_xpath_query: str = "{0}metadata/{0}title"
    severity_xpath_query: str = "{0}metadata/{0}advisory/{0}severity"
    description_xpath_query: str = "{0}metadata/{0}description"
    cve_xpath_query: str = "{0}metadata/{0}advisory/{0}cve"
    criteria_xpath_query: str = "{0}criteria"


@dataclass
class Parsed(ABC):
    identity: str


@dataclass
class Version(Parsed):
    operation: str
    value: str


@dataclass
class Artifact(Parsed):
    name: str


@dataclass
class Test(Parsed):
    artifact_id: str
    version_id: str


@dataclass()
class Impact:
    namespace_test_id: str
    affected_test_ids: list[str]


class OVALElementParser(ABC):
    oval_element = None

    @staticmethod
    @abstractmethod
    def parse(xml_element: ET.Element, config: OVALParserConfig) -> Parsed | None:
        ...

    @staticmethod
    def _find_with_regex(data: str, regex: re.Pattern):
        """
        Utility function for executing a regular expression over raw data and returning the 1st group if a match is found.
        Returns the raw data as it is if no match was found
        """
        regex_match = re.search(regex, data)
        if regex_match and len(regex_match.groups()) > 0:
            return regex_match.group(1)
        else:
            return data


class VulnerabilityParser(OVALElementParser, ABC):
    """
    Parser for processing a definition element of the definitions OVAL collection. The main parse() is intentionally
    left as an abstract function for the driver to implement since each publisher expresses the vulnerability differently
    """

    oval_element = OVALElementEnum.VULNERABILITY

    @staticmethod
    def _parse_criteria(xml_element: ET.Element, oval_ns: str, config: OVALParserConfig) -> list[Impact]:
        """
        The first or outermost criteria with in a definition, may contain nested groups of criteria/criterion bound by
        different operators

        Example of xml element parsed
          <criteria operator="OR">
            <criteria operator="AND">
              <criteria operator="OR">
                <criterion test_ref="oval:org.opensuse.security:tst:2009301949" comment="SUSE Linux Enterprise Module for Basesystem 15 SP3 is installed"/>
                <criterion test_ref="oval:org.opensuse.security:tst:2009341508" comment="SUSE Linux Enterprise Server 15 SP1-LTSS is installed"/>
              </criteria>
              <criteria operator="OR">
                <criterion test_ref="oval:org.opensuse.security:tst:2009624165" comment="sqlite3-3.36.0-3.12.1 is installed"/>
                <criterion test_ref="oval:org.opensuse.security:tst:2009624166" comment="sqlite3-devel-3.36.0-3.12.1 is installed"/>
              </criteria>
            </criteria>
            <criteria operator="AND">
              <criterion test_ref="oval:org.opensuse.security:tst:2009341508" comment="SUSE Linux Enterprise Server 15 SP2 is installed"/>
              <criterion test_ref="oval:org.opensuse.security:tst:2009624165" comment="sqlite3-3.36.0-3.12.1 is installed"/>
            </criteria>
          </criteria
        """
        criteria_element = xml_element.find(config.criteria_xpath_query.format(oval_ns))
        results = []

        if not criteria_element:
            return results

        if criteria_element.attrib["operator"].lower() == "or":
            # indicates multiple groups of impacted artifacts, parse each group and gather results
            for child in criteria_element:
                results.extend(VulnerabilityParser._parse_group(child, config))
        else:
            # indicates a single group of impacted artifacts
            results.extend(VulnerabilityParser._parse_group(criteria_element, config))

        return results

    @staticmethod
    def _parse_group(criteria_element: ET.Element, config: OVALParserConfig) -> list[Impact]:
        """
        A logical group is bunch of conditions bound by AND operator

        Example of xml element parsed
        <criteria operator="AND">
          <criteria operator="OR">
            <criterion test_ref="oval:org.opensuse.security:tst:2009301949" comment="SUSE Linux Enterprise Module for Basesystem 15 SP3 is installed"/>
            <criterion test_ref="oval:org.opensuse.security:tst:2009341508" comment="SUSE Linux Enterprise Server 15 SP1-LTSS is installed"/>
          </criteria>
          <criteria operator="OR">
            <criterion test_ref="oval:org.opensuse.security:tst:2009624165" comment="sqlite3-3.36.0-3.12.1 is installed"/>
            <criterion test_ref="oval:org.opensuse.security:tst:2009624166" comment="sqlite3-devel-3.36.0-3.12.1 is installed"/>
          </criteria>
        </criteria>
        """
        results = []

        # further parsing makes the assumption that this element has 2 children, bail out of here if that's not true
        if len(criteria_element) != 2:
            return results

        namespace_ids = VulnerabilityParser._parse_sub_group(criteria_element[0], config, config.platform_regex)

        # assumes 1st child contains platform info and 2nd contains artifact info
        if not namespace_ids:
            return results

        try:
            test_ids = VulnerabilityParser._parse_sub_group(criteria_element[1], config, config.artifact_regex)
        except:
            return results

        if not test_ids:
            return results

        for item in namespace_ids:
            results.append(
                Impact(
                    namespace_test_id=item,
                    affected_test_ids=test_ids,
                )
            )

        return results

    @staticmethod
    def _parse_sub_group(crit_element: ET.Element, config: OVALParserConfig, regex: re.Pattern) -> list[str]:
        """
        A logical sub-group is bunch of conditions bound by OR operator or a single condition

        Examples of xml element parsed
        <criteria operator="OR">
          <criterion test_ref="oval:org.opensuse.security:tst:2009301949" comment="SUSE Linux Enterprise Module for Basesystem 15 SP3 is installed"/>
          <criterion test_ref="oval:org.opensuse.security:tst:2009341508" comment="SUSE Linux Enterprise Server 15 SP1-LTSS is installed"/>
        </criteria>

        <criterion test_ref="oval:org.opensuse.security:tst:2009624165" comment="sqlite3-3.36.0-3.12.1 is installed"/>

        """
        test_ids = []
        crit_tag = OVALElementParser._find_with_regex(crit_element.tag, config.tag_regex)

        if crit_tag == "criterion":
            regex_match = re.search(regex, crit_element.attrib["comment"])
            if regex_match and crit_element.attrib["test_ref"]:
                test_ids.append(crit_element.attrib["test_ref"])
        elif crit_tag == "criteria":
            for criterion in crit_element:
                regex_match = re.search(regex, criterion.attrib["comment"])
                if regex_match and criterion.attrib["test_ref"]:
                    test_ids.append(criterion.attrib["test_ref"])

        return test_ids


class TestParser(OVALElementParser):
    """
    Parser for processing a rpminfo_test element of the tests OVAL collection
    """

    oval_element = OVALElementEnum.TEST

    @staticmethod
    def parse(xml_element: ET.Element, config: OVALParserConfig) -> Test | None:
        """
        Examples of xml elements parsed

        <rpminfo_test id="oval:org.opensuse.security:tst:2009223735" version="1" comment="sle-module-basesystem-release is ==15" check="at least one" xmlns="http://oval.mitre.org/XMLSchema/oval-definitions-5#linux">
          <object object_ref="oval:org.opensuse.security:obj:2009042619"/>
          <state state_ref="oval:org.opensuse.security:ste:2009061809"/>
        </rpminfo_test>
        """
        logger = logging.getLogger("oval-v2-parser")

        identity = artifact_id = version_id = None
        try:
            identity = xml_element.attrib["id"]
            for child in xml_element:
                child_tag = OVALElementParser._find_with_regex(child.tag, config.tag_regex)
                if child_tag == "object":
                    artifact_id = child.attrib["object_ref"]
                elif child_tag == "state":
                    version_id = child.attrib["state_ref"]
        except Exception:
            logger.exception("ignoring error parsing test xml element")
            identity = artifact_id = version_id = None

        if identity and artifact_id and version_id:
            return Test(identity=identity, artifact_id=artifact_id, version_id=version_id)
        else:
            return None


class ArtifactParser(OVALElementParser):
    """
    Parser for processing a rpminfo_object element of the objects OVAL collection
    """

    oval_element = OVALElementEnum.ARTIFACT

    @staticmethod
    def parse(xml_element: ET.Element, config: OVALParserConfig) -> Artifact | None:
        """
        Example of xml element parsed

        <rpminfo_object id="oval:org.opensuse.security:obj:2009041419" version="1" xmlns="http://oval.mitre.org/XMLSchema/oval-definitions-5#linux">
          <name>policycoreutils-python</name>
        </rpminfo_object>
        """
        logger = logging.getLogger("oval-v2-parser")

        identity = name = None
        try:
            identity = xml_element.attrib["id"]
            for child in xml_element:
                child_tag = OVALElementParser._find_with_regex(child.tag, config.tag_regex)

                if child_tag in ["name"]:
                    name = child.text
                    break
        except Exception:
            logger.exception("ignoring error parsing object xml element")
            identity = name = None

        if identity and name:
            return Artifact(identity=identity, name=name)
        else:
            return None


class VersionParser(OVALElementParser):
    """
    Parser for processing a rpminfo_state element of the state OVAL collection
    """

    oval_element = OVALElementEnum.VERSION

    @staticmethod
    def parse(xml_element: ET.Element, config: OVALParserConfig) -> Version | None:
        """
        Examples of xml elements parsed

        <rpminfo_state id="oval:org.opensuse.security:ste:2009079458" version="1" xmlns="http://oval.mitre.org/XMLSchema/oval-definitions-5#linux">
          <version operation="equals">0</version>
        </rpminfo_state>

        <rpminfo_state id="oval:org.opensuse.security:ste:2009111500" version="1" xmlns="http://oval.mitre.org/XMLSchema/oval-definitions-5#linux">
          <evr datatype="evr_string" operation="less than">0:1.15.2-4.25</evr>
          <arch datatype="string" operation="pattern match">(noarch)</arch>
        </rpminfo_state>
        """
        logger = logging.getLogger("oval-v2-parser")

        identity = op = value = None
        try:
            identity = xml_element.attrib["id"]
            for child in xml_element:
                child_tag = OVALElementParser._find_with_regex(child.tag, config.tag_regex)
                if child_tag in ["version", "evr"]:
                    op = child.attrib["operation"]
                    value = child.text
                    break
        except Exception:
            logger.exception("ignoring error parsing state xml element")
            identity = op = value = None

        if identity and op and value:
            return Version(identity=identity, operation=op, value=value)
        else:
            return None


class OVALParserFactory:
    """
    A factory that outputs a parser based on the xml content to be parsed

    Initialize the factory with a list of parser classes and an oval element enumeration.
    The latter ties a parser to the xml tag of element it can process
    """

    def __init__(self, parsers: list[type[OVALElementParser]], element_enum: type[enum.Enum]):
        if not parsers or not isinstance(parsers, list) or any(not issubclass(item, OVALElementParser) for item in parsers):
            raise ValueError("Invalid input for parsers, must be a list of OVALElementParser sub-classes")

        if not element_enum or not issubclass(element_enum, enum.Enum):
            raise ValueError("Invalid input for oval element enumeration, must be a python enum class")

        self.parser_map = {parser.oval_element: parser for parser in parsers}
        self.element_enum = element_enum

        if set(self.element_enum) != set(self.parser_map.keys()):
            raise ValueError("Parsers are not a match for the oval element enumeration")

    def get_parser(self, oval_element: enum.Enum) -> type[OVALElementParser] | None:
        """
        Returns the parser for input oval element if one is available
        """
        if oval_element and isinstance(oval_element, self.element_enum):
            return self.parser_map.get(oval_element)
        else:
            return None

    def get_oval_element(self, xml_element: ET.Element, config: OVALParserConfig) -> enum.Enum | None:
        """
        Checks and returns an oval enumeration instance if the xml content is a supported OVAL element
        """
        result = None
        if not ET.iselement(xml_element):
            return result

        tag = OVALElementParser._find_with_regex(xml_element.tag, config.tag_regex)

        try:
            result = self.element_enum(tag)
        except ValueError:
            result = None

        return result


def iter_parse_vulnerability_file(
    oval_file_path: str,
    parser_config: OVALParserConfig,
    parser_factory: OVALParserFactory,
) -> defaultdict[enum.Enum, dict[str, Parsed]]:
    """
    Starting point for parsing a vulnerability class OVAL file content.
    Iteratively parses the file using the parsers supplied by the input factory.
    Does not load the entire XML file into memory which makes it slower but OVAL files can be huge.
    Returns a dictionary where each key represents an oval element and the values are the parsed instances of that element
    """
    logger = logging.getLogger("oval-v2-parser")

    logger.info(f"parsing {oval_file_path}")
    parsed_dict: defaultdict[enum.Enum, dict[str, Parsed]] = defaultdict(dict)

    if os.path.exists(oval_file_path):
        ingress = False
        opener = open

        if oval_file_path.endswith(".gz"):
            opener = gzip.open

        with opener(oval_file_path, "rb") as f:
            for event, xml_element in iterparse(f, events=("start", "end")):
                # gather definition
                if event == "start" and parser_factory.get_oval_element(xml_element, parser_config):
                    ingress = True
                elif event == "end":
                    # is this an interesting oval element?
                    oval_element = parser_factory.get_oval_element(xml_element, parser_config)

                    # is the interesting oval element in ingress?
                    if oval_element and ingress:
                        # yes and yes, halt ingress and parse the element
                        ingress = False
                        parser = parser_factory.get_parser(oval_element)
                        if parser:
                            result = parser.parse(xml_element, parser_config)
                            if result:
                                parsed_dict[oval_element][result.identity] = result
                            else:
                                logger.warning("unable to parse %s element", repr(oval_element.value))
                        else:
                            logger.warning(
                                "no parser found for oval element %s, skipping",
                                oval_element,
                            )

                # clear the element if doesn't need to be processed or done processing
                if not ingress and event == "end":
                    xml_element.clear()
    else:
        logger.warning(f"{oval_file_path} not found, returning empty results")

    return parsed_dict



# flake8: noqa
from __future__ import annotations

import logging
import os
import re
import shlex
import shutil
import subprocess  # nosec
import tempfile
from dataclasses import dataclass

from vunnel import utils


@dataclass
class GitCommitSummary:
    sha: str
    updated: dict[str, str]
    deleted: dict[str, str]


@dataclass
class GitRevision:
    sha: str
    file: str


class UbuntuGitServer503Error(Exception):
    """Exception raised when the ubuntu git server returns a 503"""

    def __init__(self):
        super().__init__(
            "The ubuntu git server is unavailable, try again later or switch to the git protocol endpoint git://git.launchpad.net/ubuntu-cve-tracker"
        )


class GitWrapper:
    __active_retired_filename_regex__ = re.compile(r"(active|retired)/CVE-\S+")
    __cve_id_regex__ = re.compile(r"CVE-\S+")
    _check_cmd_ = "git --version"
    _is_git_repo_cmd_ = "git rev-parse --is-inside-work-tree"
    _set_remote_cmd_ = "git remote set-url origin {src}"
    _clone_cmd_ = "git clone -b {branch} {src} {dest}"
    _check_out_cmd_ = "git checkout {branch}"
    _pull_cmd_ = "git pull -f"
    _fetch_cmd_ = "git fetch --all"
    _clean_cmd_ = "git clean --force -d"
    _reset_cmd_ = "git reset --hard HEAD"
    _pull_ff_only_cmd_ = "git pull --ff-only"
    _write_graph_ = "git commit-graph write --reachable --changed-paths"
    _change_set_cmd_ = "git log --no-renames --no-merges --name-status --format=oneline {from_rev}..{to_rev}"
    _get_rev_content_cmd_ = "git show {sha}:{file}"
    _head_rev_cmd_ = "git rev-parse HEAD"
    _ubuntu_server_503_message = "error: RPC failed; HTTP 503 curl 22 The requested URL returned error: 503"

    def __init__(
        self, source: str, branch: str, checkout_dest: str, workspace: str | None = None, logger: logging.Logger | None = None
    ):
        self.src = source
        self.branch = branch
        self.dest = checkout_dest
        self.workspace = workspace if workspace else tempfile.gettempdir()

        if not logger:
            logger = logging.getLogger(self.__class__.__name__)
        self.logger = logger
        self.cve_rev_history: dict[str, list[GitRevision]] = {}

        try:
            out = self._exec_cmd(self._check_cmd_)
            self.logger.trace("git executable verified using cmd: {}, output: {}".format(self._check_cmd_, out.decode()))
        except:
            self.logger.exception('could not find required "git" executable. Please install git on host')
            raise

    def _check(self, destination):
        try:
            if not os.path.exists(destination):
                self.logger.debug(f"git working tree not found at {destination}")
                return False

            cmd = self._is_git_repo_cmd_
            out = self._exec_cmd(cmd, cwd=destination)
            self.logger.debug("check for git repository, cmd: {}, output: {}".format(cmd, out.decode()))
        except:
            self.logger.debug(f"git working tree not found at {destination}")
            return False

        return True

    def _delete_repo(self):
        if os.path.exists(self.dest):
            self.logger.debug("deleting existing repository")
            shutil.rmtree(self.dest, ignore_errors=True)

    def _clone_repo(self):
        try:
            self.logger.info(f"cloning git repository {self.src} branch {self.branch} to {self.dest}")
            cmd = self._clone_cmd_.format(src=self.src, dest=self.dest, branch=self.branch)
            out = self._exec_cmd(cmd)
            self.logger.debug("initialized git repo, cmd: {}, output: {}".format(cmd, out.decode()))
            self._write_graph()
        except:
            self.logger.exception(f"failed to clone git repository {self.src} branch {self.branch} to {self.dest}")
            raise

    @utils.retry_with_backoff()
    def init_repo(self, force=False):
        if force:
            self._delete_repo()

        if self._check(self.dest):
            self.logger.debug("found git repository at {}".format(self.dest))
            self.sync_with_upstream()
            return

        self._clone_repo()

    def parse_full_cve_revision_history(self, git_log_output: str) -> dict[str, list[GitRevision]]:
        hist = {}
        entries = self._parse_log(git_log_output)

        for entry in entries:
            for cve, file in entry.updated.items():
                if cve not in hist:
                    hist[cve] = []

                hist[cve].append(GitRevision(entry.sha, file))

        return hist

    def prepare_cve_revision_history(self):
        self.logger.info("building full revision history for all CVEs.  This may take quite some time.")
        self.cve_rev_history = {}
        out = self._exec_cmd("git log --name-status --no-merges --format=oneline -- retired/ active/", cwd=self.dest)
        self.cve_rev_history = self.parse_full_cve_revision_history(out.decode())

    @utils.retry_with_backoff()
    def sync_with_upstream(self):
        try:
            try:
                self._exec_cmd(self._set_remote_cmd_.format(src=self.src), cwd=self.dest)

                # Cleanup any untracked files which might be present and reset any changes on the current branch
                try:
                    self._exec_cmd(self._clean_cmd_, cwd=self.dest)
                    self._exec_cmd(self._reset_cmd_, cwd=self.dest)
                except:
                    pass

                self._exec_cmd(self._check_out_cmd_.format(branch=self.branch), cwd=self.dest)
            except:  # nosec
                pass

            try:
                out = self._exec_cmd(self._pull_ff_only_cmd_, cwd=self.dest)
                self.logger.debug("synced with upstream git repo, output: {}".format(out.decode()))
                self._write_graph()
            except UbuntuGitServer503Error:
                raise
            except:  # nosec
                # if something other than 503 occurred at this point just remove the repo and re-clone
                self._delete_repo()
                self._clone_repo()

        except:
            self.logger.exception("failed to git pull")
            raise

    def _write_graph(self):
        """
        writes out a binary representation of the commit graph as a git object, enabling
        huge performance gains when traversing the graph (e.g. git log)
        """
        try:
            self.logger.debug("writing commit graph")
            self._exec_cmd(self._write_graph_, cwd=self.dest)
        except:
            self.logger.exception("failed to write commit graph")
            raise

    def get_merged_change_set(self, from_rev: str, to_rev: str | None = None) -> tuple[dict[str, str], dict[str, str]]:
        try:
            self.logger.trace("fetching changes set between revisions {} - {}".format(from_rev, to_rev))
            cmd = self._change_set_cmd_.format(from_rev=from_rev, to_rev=to_rev if to_rev else "")
            out = self._exec_cmd(cmd, cwd=self.dest)
            commit_list = self._parse_log(out.decode())
            return self._compute_change_set(commit_list)
        except:
            self.logger.exception("failed to compute logically modified and removed CVEs between commits")
            raise

    def get_revision_history(self, cve_id: str, file_path: str, from_rev: str | None = None) -> list[GitRevision]:
        try:
            self.logger.trace("fetching revision history for {}".format(file_path))
            return self.cve_rev_history.get(cve_id, [])
        except:
            self.logger.exception("failed to fetch the revision history for {}".format(file_path))
            raise

    def get_content(self, git_rev: GitRevision) -> list[str]:
        if not isinstance(git_rev, GitRevision):
            raise ValueError("Input must be a GitRevision")

        try:
            cmd = self._get_rev_content_cmd_.format(sha=git_rev.sha, file=git_rev.file)
            out = self._exec_cmd(cmd, cwd=self.dest)
            return out.decode().splitlines()
        except:
            self.logger.exception("failed to get content for {} from git commit {}".format(git_rev.file, git_rev.sha))

    def get_current_rev(self) -> str:
        try:
            rev = self._exec_cmd(self._head_rev_cmd_, cwd=self.dest)
            return rev.decode().strip() if isinstance(rev, bytes) else rev
        except:
            self.logger.exception("unable to get current git revision")

    @staticmethod
    def _parse_revision(rev_raw: list[str]) -> GitRevision:
        """
        List containing two items
        [
            'eabaf525ae78eea3cd9f6063721afd1111efcd5c ran process_cves',
            'R100    active/CVE-2017-16011   ignored/CVE-2017-16011'
        ]
        :param rev_raw:
        :return:
        """
        commit_line_components = rev_raw[0].split()
        change_line_components = rev_raw[1].split()

        return GitRevision(
            sha=commit_line_components[0],
            status=change_line_components[0],
            file=change_line_components[2] if len(change_line_components) >= 3 else change_line_components[1],
        )

    @staticmethod
    def _compute_change_set(commit_list: list[GitCommitSummary]) -> tuple[dict[str, str], dict[str, str]]:
        """
        List of GitCommitSummary tuples in the log order (last commit first)
        [
            GitCommitSummary(sha='9d7608a304733e936771d53d66c17ea1ed81ca2a', updated={'CVE-2018-1108': 'active/CVE-2018-1108'}, deleted={})
            GitCommitSummary(sha='34cc8435c53610f4a0ca5ff355151eb5a6af0df9', updated={'CVE-2018-12714': 'active/CVE-2018-12714'}, deleted={})
        ]


        :param commit_list:
        :return:
        """
        # reverse the commit list and process commits in the order they were inserted
        commit_list.reverse()

        modified: dict[str, str] = {}
        removed: dict[str, str] = {}

        for commit in commit_list:
            # Overlay each commit on top of computed results

            # start logically deleted items in the commit and filter it out from from the overall modified result
            modified = {key: value for key, value in modified.items() if key not in commit.deleted}

            # merge updated items in the commit with the overall modified result
            modified.update(commit.updated)

            # merge deleted items in the commit with the overall removed result
            removed.update(commit.deleted)

        # filter out anything from the overall removed result to handle anything was logically removed and re-added
        removed = {key: value for key, value in removed.items() if key not in modified}

        return modified, removed

    def _parse_log(self, ordered_changes: str) -> list[GitCommitSummary]:
        """
        Input in the form

        f9cc163f61a86209ca955017c56c1eedd285bcee html-export: convert from BZR_COMMIT to GIT_COMMIT
        M       Makefile.include
        M       ignored/Makefile
        M       scripts/generate-pkgs-makefile.py
        M       scripts/html-report
        6571b3166f2a79799d3bff89b9eb319b720812ac generate-graph.py: fix chdir breakage when in $UCT
        M       scripts/generate-graphs.py
        66fdf2b24c2151e0d36a7e9a1aea9714ed7ad13e generate-graphs.py: parse git log as part of UCT git conversion
        M       scripts/generate-graphs.py
        53eeed7959e55da1854d3f9627b9652d93ef371f reports; convert from bzr to git
        M       scripts/monthly-report
        M       scripts/yearly-report
        2dfafda744cf2d7da70d1d0598c2fe1ede1955fd git conversion: move .bzrignore to .gitignore
        D       .bzrignore
        A       .gitignore

        :param ordered_changes:
        :return:
        """

        commits_list: list[GitCommitSummary] = []
        commit_lines: list[list[str]] = []
        # split lines and remove any empty string in between or at the end
        ordered_changes_iterator = (item.strip() for item in ordered_changes.splitlines() if item.strip())

        for line in ordered_changes_iterator:
            components = line.split()
            if components and len(components) > 1:
                if len(components[0]) > 5:  # indicates this is a commit sha since length greater than any change status
                    if commit_lines:  # encountered next commit, process the stored one first
                        c = self._parse_normalized_commit(commit_lines)
                        commits_list.append(c) if c else None
                        del commit_lines[:]
                    # else:  # encountered the first commit line, keep going
                    #     pass

                    commit_lines.append(components)
                else:
                    commit_lines.append(components)
            else:
                pass

        # process the last commit if any
        if commit_lines:
            c = self._parse_normalized_commit(commit_lines)
            commits_list.append(c) if c else None
            del commit_lines[:]

        return commits_list

    def _parse_normalized_commit(self, commit_lines: list[list[str]]) -> GitCommitSummary | None:
        """
        A list of lists where each inner list represents a line in the commit log
        [
            ['f1743ee4e985c0c3cf6dd8ccaa0fdfd801415f0d', 'Process', 'cves', 'run:', 'triaged', '8', 'CVEs,', '40', 'Ignored,', '4', 'Packages'],
            ['A', 'active/CVE-2018-1000201'],
            ['A', 'active/CVE-2018-12641'],
            ['A', 'active/CVE-2018-12689'],
            ['A', 'active/CVE-2018-12697'],
            ['A', 'active/CVE-2018-12698'],
            ['A', 'active/CVE-2018-12699'],
            ['A', 'active/CVE-2018-12700'],
            ['A', 'active/CVE-2018-12713'],
            ['M', 'check-cves.log'],
            ['M', 'ignored/not-for-us.txt']
        ]

        :param commit_lines:
        :return:
        """
        updated: dict[str, str] = {}
        deleted: dict[str, str] = {}

        for components in commit_lines[1:]:
            if re.match(self.__active_retired_filename_regex__, components[1]):
                cve_id = components[1].split("/", 1)[1]
                if components[0] == "A" or components[0] == "M":
                    # CVE added to or modified in active or retired directory
                    # A       retired/CVE-2018-7332
                    # M       retired/CVE-2018-7480
                    updated[cve_id] = components[1]
                elif components[0] == "D":
                    # CVE deleted from active or retired directory
                    # D       active/CVE-2009-1553
                    deleted[cve_id] = components[1]
                elif components[0].startswith("R"):
                    if components[0] != "R100" and len(components) > 2:
                        updated[cve_id] = components[2]
                else:
                    # either not a commit line or an irrelevant file, ignore it
                    self.logger.debug("skipping unknown change symbol {}".format(components[0]))
            else:
                # not a match
                pass

        if updated or deleted:
            deleted = {key: value for key, value in deleted.items() if key not in updated}
            return GitCommitSummary(sha=commit_lines[0][0], updated=updated, deleted=deleted)
        else:
            return None

    def _exec_cmd(self, cmd, *args, **kwargs) -> bytes:
        """
        Run a command with errors etc handled
        :param cmd: list of arguments (including command name, e.g. ['ls', '-l])
        :param args:
        :param kwargs:
        :return:
        """
        try:
            self.logger.trace("running: {}".format(cmd))
            cmd_list = shlex.split(cmd)
            return subprocess.check_output(cmd_list, *args, **kwargs, stderr=subprocess.PIPE)  # nosec
        except Exception as e:
            self.logger.exception("error executing command: {}".format(cmd))

            if isinstance(e, subprocess.CalledProcessError):
                if e.stderr and self._ubuntu_server_503_message in e.stderr.decode():
                    raise UbuntuGitServer503Error()

            raise e
